@@ -152,6 +152,14 @@ typedef struct {
 	/* If true, we are doing extended query message */
 	bool doing_extended_query_message;
 
+	/* If true, we have rewritten where_to_send map in the current query
+	 * context. pool_unset_query_in_progress() should restore the data from
+	 * where_to_send_save.  For now, this is only necessary while doing
+	 * extended query protocol and in streaming replication mode.
+	 */
+	bool need_to_restore_where_to_send;
+	bool where_to_send_save[MAX_NUM_BACKENDS];
+
 	/* If true, the command in progress has finished successfully. */
 	bool command_success;
 
@@ -291,6 +299,8 @@ extern void pool_pending_message_reset_previous_message(void);
 extern void pool_pending_message_set_previous_message(POOL_PENDING_MESSAGE *message);
 extern POOL_PENDING_MESSAGE *pool_pending_message_get_previous_message(void);
 extern bool pool_pending_message_exists(void);
+extern const char *pool_pending_message_type_to_string(POOL_MESSAGE_TYPE type);
+extern POOL_PENDING_MESSAGE *pool_pending_message_find_lastest_by_query_context(POOL_QUERY_CONTEXT *qc);
 extern void dump_pending_message(void);
 extern void pool_set_major_version(int major);
 extern void pool_set_minor_version(int minor);
